'use strict';
angular.module('aom.services')
  .factory('aomOfflineDatabasesSrv', function ($state, $ionicPlatform, $cordovaSQLite, $q, $timeout, $ionicHistory, aomPopupSrv, aomAPISrv, aomDatabaseSrv, aomResourcesSrv, aomNotificationSrv, aomLogSrv, aomNetworkSrv, constants) {
    var aomOfflineDatabases = {};

    aomOfflineDatabases.data = {
      databases: [],
      selectedOfflineDatabase: {},
      importing: false,
      importFailed: false
    };

    function openCordovaCustomerDb() {
      $ionicPlatform.ready(function () {
        aomOfflineDatabases.customerdb = $cordovaSQLite.openDB({ name: 'customerv5.db', location: 'default' });
      });
    }

    function openWebCustomerDb() {
      aomOfflineDatabases.customerdb = window.openDatabase('customerv5.db', '1.0', 'Customer', -1);
    }

    function setupCustomerDb() {
      window.cordova ? openCordovaCustomerDb() : openWebCustomerDb();
    }

    function notifyOnOfflineDatabaseScreen(message) {
      // if ($ionicHistory.currentStateName() === 'app.OfflineDatabases') {
        aomNotificationSrv.notify(message);
      // }
    }

    function runQueryOnCustomerDatabase(query) {
      return $cordovaSQLite.execute(aomOfflineDatabases.customerdb, query, []);
    }

    function dropTableFromCustomerDatabase(table) {
      return runQueryOnCustomerDatabase('DROP TABLE IF EXISTS [' + table + '];');
    }

    function removeOfflineDatabase(name) {
      var unavailableName = constants.offlineDatabases.unavailable + name;
      return dropTableFromCustomerDatabase(name)
        .then(function () {
          return dropTableFromCustomerDatabase(unavailableName);
        });
    }

    function importDBOnWindows(statements, statementId) {
      return runQueryOnCustomerDatabase(statements[statementId])
        .then(function () {
          if (statementId !== statements.length - 1) {
            return importDBOnWindows(statements, statementId + 1);
          } else {
            return $q.when();
          }
        });
    }

    function importSql(response) {
      var deferred = $q.defer();

      var sql = response.replace('BEGIN TRANSACTION;', '')
        .replace('COMMIT TRANSACTION;', '')
        .replace(/;INSERT INTO/g, ';@#&123INSERT INTO')
        .replace(/;CREATE TABLE/g, ';@#&123CREATE TABLE')
        .split('@#&123');

      if (ionic.Platform.isEdge() || !window.cordova) {
        return importDBOnWindows(sql, 0);
      } else {
        aomOfflineDatabases.customerdb.sqlBatch(sql, function () {
          deferred.resolve();
        }, function (error) {
          deferred.reject(error);
        });
      }
      return deferred.promise;
    }

    function handleLongRunningSql() {
      if (ionic.Platform.isIOS()) {
        aomOfflineDatabases.customerdb.interruptSQL();
        $state.go('app.Jobs');
        aomNotificationSrv.notify(aomResourcesSrv.data.resources.Job.LongRunningSqlAlertMessageiOS, false, { name: 'LongRunningSQL', maxFreq: 20000 });
      } else {
        aomPopupSrv.alert('longRunningSqlNotification', {
          title: aomResourcesSrv.data.resources.Job.LongRunningSql,
          template: aomResourcesSrv.data.resources.Job.LongRunningSqlAlertMessage,
          onDismiss: function () {
            aomLogSrv.log({ message: aomResourcesSrv.data.resources.Job.LongRunningSql }, aomLogSrv.levels.medium, aomLogSrv.severities.err);
            if (ionic.Platform.isAndroid()) {
              ionic.Platform.exitApp();
            } else if (ionic.Platform.isEdge()) {
              WinJS.Application.stop();
            }
          }
        });
      }
    }

    function countRows(name) {
      return runQueryOnCustomerDatabase('SELECT count(*) as myCount FROM [' + name + '];')
        .then(function (res) {
          return $q.when(res.rows.item(0).myCount);
        });
    }

    function deleteAllCustomerDbData() {
      return getListOfAllOfflineDatabases()
        .then(function (databases) {
          var promises = databases.map(function (database) {
            return dropTableFromCustomerDatabase(database);
          });
          return $q.all(promises);
        })
        .then(function () {
          return aomDatabaseSrv.deleteQuery({ OfflineDatabases: {} });
        });
    }

    function getListOfAllOfflineDatabases() {
      return runQueryOnCustomerDatabase('SELECT name FROM sqlite_master WHERE type=\'table\';')
        .then(function (response) {
          var databases = [];
          for (var i = 0; i < response.rows.length; i++) {
            databases.push(response.rows.item(i).name);
          }
          return $q.when(databases);
        });
    }

    function offlineDatabaseAccess(name, options) {
      var unavailableName = constants.offlineDatabases.unavailable + name;
      var oldName = options ? unavailableName : name;
      var newName = options ? name : unavailableName;
      var statement = 'ALTER TABLE [' + oldName + '] RENAME TO [' + newName + '];';

      return runQueryOnCustomerDatabase(statement)
        .catch(function (err) {
          aomLogSrv.log({ message: 'Could not rename ' + oldName + 'to unavailable_', error: err }, aomLogSrv.levels.medium, aomLogSrv.severities.warn);
          return $q.when();
        });
    }

    function insertOfflineDatabaseRecord(database) {
      return aomDatabaseSrv.insertQuery({ OfflineDatabases: { id: database.Id, database_json: database } }, true, false);
    }

    function deleteOfflineDatabaseRecord(database) {
      return removeOfflineDatabase(database.Name)
        .then(function () {
          return aomDatabaseSrv.deleteQuery({ OfflineDatabases: { id: database.Id } });
        });
    }

    function updateNamesOfflineDatabasesAlreadyDownloaded(inactiveDatabases, activeDatabases, removeDatabases) {
      return executeOperation(deleteOfflineDatabaseRecord, removeDatabases)
        .then(function () {
          return executeOperation(offlineDatabaseAccess, inactiveDatabases, false);
        })
        .then(function () {
          return executeOperation(offlineDatabaseAccess, activeDatabases, true);
        });
    }

    function downloadAndImportOfflineDatabase(database) {
      return aomAPISrv.getData(database.links.self.href)
        .then(function (response) {
          return importSql(response);
        })
        .then(function () {
          return countRows(database.Name);
        })
        .then(function (rows) {
          database.status = true;
          database.Rows = rows;
          aomOfflineDatabases.data.importFailed = false;
        })
        .catch(function (err) {
          database.status = false;
          aomOfflineDatabases.data.importFailed = true;
          aomLogSrv.log({ message: 'Failed to Import SQL', error: err }, aomLogSrv.levels.low, aomLogSrv.severities.warn);
          return $q.when();
        })
        .finally(function () {
          return insertOfflineDatabaseRecord(database);
        });
    }

    function executeOperation(operation, list, options) {
      var promises = list.map(function (database) {
        return operation(database, options);
      });
      return $q.all(promises);
    }

    function updateObjectFromManifest(response) {
      var final = [];
      var existing = aomOfflineDatabases.data.databases;
      var manifest = response.DBs;
      var manifestEnumerable = Enumerable.from(manifest);

      var inactiveDatabases = [];
      var activeDatabases = [];
      var removeDatabases = [];

      for (var i = 0; i < existing.length; i++) {
        var pushObj = existing[i];
        var newManifestVersion = manifestEnumerable.where('$.Id !== "' + existing[i].Id + '" && $.Name === "' + existing[i].Name + '"').toArray();
        var manifestVersion = manifestEnumerable.where('$.Id === "' + existing[i].Id + '"').toArray();
        var doesSameDatabaseNameExist = manifestEnumerable.where('$.Name === "' + existing[i].Name + '"').toArray();
        var active = existing[i].active;

        existing[i].active = doesSameDatabaseNameExist.length === 1 && manifestVersion.length === 1;

        if (active !== undefined) {
          if (active && !existing[i].active) {
            inactiveDatabases.push(pushObj.Name);
          } else if (!active && existing[i].active) {
            activeDatabases.push(pushObj.Name);
          }
        }

        if (manifestVersion.length === 1) {
          manifestVersion[0].lastSeen = moment().format();
          manifestVersion[0].added = true;
        } else if (newManifestVersion.length === 1) {
          newManifestVersion[0].lastSeen = moment().format();
          newManifestVersion[0].active = true;
          pushObj = newManifestVersion[0];
          newManifestVersion[0].added = true;
        }
        final.push(pushObj);

        if (pushObj.Name === existing[i].Name && pushObj.Id !== existing[i].Id) {
          removeDatabases.push({ Id: existing[i].Id, Name: existing[i].Name });
        }
      }

      var beforeFinal = manifestEnumerable.where('$.added === undefined').toArray();
      for (var x = 0; x < beforeFinal.length; x++) {
        beforeFinal[x].lastSeen = moment().format();
        beforeFinal[x].active = true;
        final.push(beforeFinal[x]);
      }
      return $q.when({ final: final, active: activeDatabases, inactive: inactiveDatabases, remove: removeDatabases });
    }

    function getDatabases() {
      var response;
      return aomAPISrv.getData(aomAPISrv.data.rootLinks.OfflineDatabases.href)
        .then(function (response) {
          return updateObjectFromManifest(response);
        })
        .then(function (finalObj) {
          response = finalObj.final;
          return updateNamesOfflineDatabasesAlreadyDownloaded(finalObj.inactive, finalObj.active, finalObj.remove);
        })
        .then(function () {
          return aomDatabaseSrv.deleteQuery({ OfflineDatabases: {} });
        })
        .then(function () {
          return executeOperation(insertOfflineDatabaseRecord, response);
        })
        .then(function () {
          return response;
        });
    }

    function startImport(final) {
      return executeOperation(downloadAndImportOfflineDatabase, Enumerable.from(final).where('$.status !== true').toArray())
        .then(function () {
          if (!aomOfflineDatabases.data.importFailed) {
            aomNotificationSrv.notify(aomResourcesSrv.data.resources.OfflineDatabases.DatabasesDownloadSuccess);
          }
          return $q.when();
        })
        .finally(function () {
          if (aomOfflineDatabases.data.importFailed) {
            aomNotificationSrv.notify(aomResourcesSrv.data.resources.OfflineDatabases.DatabasesDownloadFailure);
            aomLogSrv.log({ message: 'Failed to import/download databases' }, aomLogSrv.levels.user, aomLogSrv.severities.err);
          }
          aomOfflineDatabases.data.importing = false;
          return aomOfflineDatabases.getStoredOfflineDatabaseRecords();
        });
    }

    function fetchOfflineDatabases() {
      aomOfflineDatabases.data.importing = true;
      return aomOfflineDatabases.getStoredOfflineDatabaseRecords()
        .then(function () {
          return getDatabases();
        })
        .then(function (final) {
          var deferred = $q.defer();
          if (Enumerable.from(final).where('$.status !== true && $.active').toArray().length > 0 && $ionicHistory.currentStateName() !== 'logon') {
            startImport(final)
              .then(function () {
                deferred.resolve();
              });
          } else {
            notifyOnOfflineDatabaseScreen(aomResourcesSrv.data.resources.OfflineDatabases.DatabasesUptoDate);
            aomOfflineDatabases.data.importing = false;
            deferred.resolve();
          }

          return deferred.promise;
        })
        .catch(function () {
          aomOfflineDatabases.data.importing = false;
          notifyOnOfflineDatabaseScreen(aomResourcesSrv.data.resources.appConfig.UseMobileNetworkForOfflineDatabases ? aomResourcesSrv.data.resources.OfflineDatabases.OfflineDatabaseOnWifiOrCellular : aomResourcesSrv.data.resources.OfflineDatabases.OfflineDatabaseOnWifi);
          return aomOfflineDatabases.getStoredOfflineDatabaseRecords();
        });
    }

    setupCustomerDb();

    aomOfflineDatabases.fetchOfflineDatabases = function () {
      if (aomNetworkSrv.onlineOnWifi() || (aomResourcesSrv.data.resources.appConfig.UseMobileNetworkForOfflineDatabases && aomNetworkSrv.onlineOnCellularNetwork())) {
        if (!aomOfflineDatabases.data.importing) {
          return fetchOfflineDatabases();
        } else {
          return $q.reject();
        }
      } else {
        notifyOnOfflineDatabaseScreen(aomResourcesSrv.data.resources.appConfig.UseMobileNetworkForOfflineDatabases ? aomResourcesSrv.data.resources.OfflineDatabases.OfflineDatabaseOnWifiOrCellular : aomResourcesSrv.data.resources.OfflineDatabases.OfflineDatabaseOnWifi);
      }
    };

    aomOfflineDatabases.getStoredOfflineDatabaseRecords = function () {
      return aomDatabaseSrv.selectQuery({ OfflineDatabases: { database_json: null } })
        .then(function (response) {
          aomOfflineDatabases.data.databases = response.rows;
          return $q.when();
        });
    };

    aomOfflineDatabases.deleteOfflineDatabases = function () {
      aomOfflineDatabases.data.importing = false;
      return deleteAllCustomerDbData()
        .then(function () {
          return aomOfflineDatabases.getStoredOfflineDatabaseRecords();
        });
    };

    aomOfflineDatabases.cleanOfflineDatabases = function () {
      var inactiveDatabases = aomOfflineDatabases.data.databases.filter(function (database) {
        return database.active === false && moment().diff(database.lastSeen, 'days') > aomResourcesSrv.data.resources.OfflineDatabases.MaximumDatabaseDays;
      });
      return executeOperation(deleteOfflineDatabaseRecord, inactiveDatabases);
    };

    aomOfflineDatabases.runQuery = function (query) {
      var defered = $q.defer(),
        result,
        prop;

      $cordovaSQLite.execute(aomOfflineDatabases.customerdb, query)
        .then(function (res) {
          if (res.rows.length > 0) {
            var sqlResultSetRowList = res.rows.item(0);
            if (res.rows.length === 1) {
              result = '';
              for (prop in sqlResultSetRowList) {
                if (sqlResultSetRowList.hasOwnProperty(prop)) {
                  result += sqlResultSetRowList[prop] + ', ';
                }
              }
              result = result.substr(0, result.length - 2);
            } else {
              result = [];
              for (var i = 0; i < res.rows.length; i++) {
                var rowResult = '';
                for (prop in sqlResultSetRowList) {
                  if (sqlResultSetRowList.hasOwnProperty(prop)) {
                    rowResult += res.rows.item(i)[prop] + ', ';
                  }
                }
                rowResult = rowResult.substr(0, rowResult.length - 2);
                result.push(rowResult);
              }
            }
            defered.resolve(result);
          } else {
            defered.resolve('null');
          }
        }, function () {
          defered.resolve('null');
        });

      return defered.promise;
    };

    return aomOfflineDatabases;
  });
